@@ -8,16 +8,16 @@
    cases as published by the Free Software Foundation.
 */
 
-#include "event.h"
+#include "gf-event.h"
 #include "rpcsvc.h"
 
 void
-rpcsvc_autoscale_threads (glusterfs_ctx_t *ctx, rpcsvc_t *rpc, int incr)
+rpcsvc_autoscale_threads(glusterfs_ctx_t *ctx, rpcsvc_t *rpc, int incr)
 {
-        struct event_pool       *pool           = ctx->event_pool;
-        int                      thread_count   = pool->eventthreadcount;
+    struct event_pool *pool = ctx->event_pool;
+    int thread_count = pool->eventthreadcount;
 
-        pool->auto_thread_count += incr;
-        (void) event_reconfigure_threads (pool, thread_count+incr);
-        rpcsvc_ownthread_reconf (rpc, pool->eventthreadcount);
+    pool->auto_thread_count += incr;
+    (void)event_reconfigure_threads(pool, thread_count + incr);
+    rpcsvc_ownthread_reconf(rpc, pool->eventthreadcount);
 }
